part of '../music_notes.dart';

/// A interface for items that can form scales.
abstract interface class Scalable<T> implements Transposable<T> {
  /// Returns the [Interval] between this [T] and [other].
  Interval interval(T other);

  /// Returns the difference in semitones between this [T] and [other].
  int difference(T other);
}

/// A Scalable iterable.
extension ScalableIterable<T extends Scalable<T>> on Iterable<T> {
  /// Returns the [Interval]s between [T]s in this [Iterable].
  Iterable<Interval> get intervalSteps sync* {
    for (var i = 0; i < length - 1; i++) {
      yield elementAt(i).interval(elementAt(i + 1));
    }
  }

  /// Returns the descending [Interval]s between [T]s this [Iterable].
  Iterable<Interval> get descendingIntervalSteps sync* {
    for (var i = 0; i < length - 1; i++) {
      yield elementAt(i + 1).interval(elementAt(i));
    }
  }

  /// Returns this [Iterable] transposed by [interval].
  Iterable<T> transposeBy(Interval interval) =>
      map((item) => item.transposeBy(interval));
}
